/*	$OpenBSD: vad.c,v 1.0 2003/04/07 06:14:47 vp Exp $	*/

/*
 * Copyright (c) 2003 Vassilis Prevelakis.  All rights reserved.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the Computer Systems
 *	Engineering Group at Lawrence Berkeley Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#define NSLAVES		1	

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/errno.h>
#include <sys/ioctl.h>
#include <sys/syslog.h>
#include <sys/device.h>
#include <sys/proc.h>
#include <sys/kernel.h>
#include <sys/fcntl.h>
#include <sys/malloc.h>

#include <machine/cpu.h>
#include <machine/intr.h>
#include <machine/bus.h>

#include <sys/audioio.h>
#include <dev/audio_if.h>
#include <dev/midi_if.h>
#include <dev/mulaw.h>
#include <dev/auconv.h>
#include <dev/audiovar.h>

#include <dev/vadio.h>

/*
#define DEFINE_TRANSLATIONS
#include <dev/isa/pasreg.h>
*/

#ifdef AUDIO_DEBUG
#define DPRINTF(x)	if (vad_debug) printf x
int	vad_debug = 0;
#else
#define DPRINTF(x)	
/* VPVP #define DPRINTF(x) */
#endif

/*
 * according to the dev/audiovar.h AU_RING_SIZE is 2^16, what fits
 * in our limits perfectly, i.e. setting it to higher value
 * in your kernel config would improve perfomance, still 2^21 is the max
 */

struct vad_softc {
	struct device sc_dev;
	audio_device_t sc_audev;
	struct audio_softc * sc_audsc; 
	
	struct encoding_info sc_ec;
	
	/* other info such as multicast address goes here */

};

/* VAD flags */
#define VAD_BUSY	0x08		/* Signal that VAD is writing       */
#define VAD_PARAM	0x10		/* Signal that audio params changed */

/* SVAD flags */
#define SVAD_BUSY       0x20


/* Special Byte Pattern to signal audio paramter changing */
char * byte_patt = "VADPARAM";


int	vad_getdev __P((void *, struct audio_device *));

int vad_open __P((void *, int));
void vad_close __P((void *));
int vad_query_encoding __P((void *, struct audio_encoding *));
int vad_set_params __P((void *, int, int, struct audio_params *,
    struct audio_params *));
int vad_round_blocksize __P((void *, int));
int vad_halt_output __P((void *));
int vad_halt_input __P((void *));
int vad_set_port __P((void *, mixer_ctrl_t *));
int vad_get_port __P((void *, mixer_ctrl_t *));
int vad_query_devinfo __P((void *, mixer_devinfo_t *));
void *vad_allocm __P((void *, int, size_t, int, int));
void vad_freem __P((void *, void *, int));
int vad_get_props __P((void *));
int vad_trigger_output __P((void *, void *, void *, int, void (*)(void *),
    void *, struct audio_params *));
int vad_trigger_input __P((void *, void *, void *, int, void (*)(void *),
    void *, struct audio_params *));



/* The softc structure for the slave vad pseudo-device */
struct svad_softc   {

	struct audio_softc * sc_audsc;
		
	struct encoding_info sc_ec;
	
	/* Status flags for VAD */
	int sc_flags;		
	
	/* Params passed to trigger_output, used in kthread */	
	void * sc_acbstart;
	void * sc_acbend;
	int sc_blksize;
	void (*sc_intrp)(void *);
	void * sc_argp;
	void * sc_outp;
	
	/* Temp Buffer for leftover block */
	int sc_blen;
	int sc_toff;
	void * sc_tbuf;
		
 	u_char sc_rchan;

} * svad_softcs = NULL; 


/* Allocates n number of svad_softcs, called by vadattach  */
void svadattach(int);



/*
 * Define our interface to the higher level audio driver.
 */

struct audio_hw_if vad_hw_if = {
	vad_open,
	vad_close,
	NULL, /* drain */
	vad_query_encoding,
	vad_set_params,
	vad_round_blocksize,
	NULL, /* commit_settings */   
	NULL, /* init_output */
	NULL, /* init_input */
	NULL, /* start_output */
	NULL, /* start_input */
	vad_halt_output,
	vad_halt_input,
	NULL, /* speaker_ctl */
	vad_getdev,
	NULL, /* setfd */
	vad_set_port,
	vad_get_port,
	vad_query_devinfo,
	vad_allocm, /* vad_allocm, */
	vad_freem,  /* freem  */ 
	NULL,  /* round */
        NULL, /* vad_mappage, */
	vad_get_props,
	vad_trigger_output, 
        vad_trigger_input
};

int	vadprobe __P((struct device *, void *, void *));
void	vadattach __P((struct device *, struct device *, void *));


struct cfattach vad_ca = {
	sizeof(struct vad_softc), vadprobe, vadattach
};

struct cfdriver vad_cd = {
	NULL, "vad", DV_DULL
};

/*
 * Probe / attach routines.
 */

/*
 * We keep a counter of the number of vads probed already, if it is greater than NVADS
 * we will return 0.
 */
int num_vads_probed = 0;
	
int
vadprobe(parent, match, aux)
	struct device *parent;
	void *match, *aux;
{

	DPRINTF(("vad%d: found \n", num_vads_probed));
	
	if(num_vads_probed < NSLAVES)  {	
		num_vads_probed++;	
		return 1;
	}
	
	else
		return 0;
	
}


/*
 * Attach hardware to driver, attach hardware driver to audio
 * pseudo-device driver .
 */
void
vadattach(parent, self, aux)
	struct device *parent, *self;
	void *aux;
{
	struct vad_softc *sc = (struct vad_softc *)self;
	
	printf(" Virtual Audio Driver %d Attached\n", sc->sc_dev.dv_unit);

	/* Attach audio MI driver to vad */
	sc->sc_audsc = (struct audio_softc *) audio_attach_mi(&vad_hw_if, sc, &sc->sc_dev);

	DPRINTF(("vadattach: audio%d attached to vad\n", sc->sc_audsc->dev.dv_unit));
	
	/* attach the number of slave devices */
	/* this will be called and unescessary amount of times, but it wont allocate */
	/* more slaves than it needs 						     */
	svadattach(NSLAVES);
	
}

int vad_open(v, flags)
	void *v;
	int flags;
{

	DPRINTF(("vad_open(flags = 0x%x)\n", flags));

	/* We need to check if the corresponding svad device has been configured */
	if(svad_softcs == NULL)  {
		DPRINTF(("vad_open: corresponding svad is not attached"));
		return(ENXIO);
	}

	return(0);
}

void vad_close(v)
        void *v;
{
	
	DPRINTF(("vad_close()\n"));
}

int vad_query_encoding(v, aep)
        void *v;
        struct audio_encoding *aep;
{
	DPRINTF(("vad_query_encoding(index=%d)\n", aep->index));
	switch (aep->index) {
	case 0:
		strlcpy(aep->name, AudioEulinear, 13);
		aep->encoding = AUDIO_ENCODING_ULINEAR;
		aep->precision = 8;
		aep->flags = 0;
		return (0);
	case 1:
		strlcpy(aep->name, AudioEmulaw,  sizeof(aep->name));
		aep->encoding = AUDIO_ENCODING_ULAW;
		aep->precision = 8;
		aep->flags = AUDIO_ENCODINGFLAG_EMULATED;
		return (0);
	case 2:
		strlcpy(aep->name, AudioEalaw, sizeof(aep->name));
		aep->encoding = AUDIO_ENCODING_ALAW;
		aep->precision = 8;
		aep->flags = AUDIO_ENCODINGFLAG_EMULATED;
		return (0);
	case 3:
		strlcpy(aep->name, AudioEslinear, sizeof(aep->name));
		aep->encoding = AUDIO_ENCODING_SLINEAR;
		aep->precision = 8;
		aep->flags = AUDIO_ENCODINGFLAG_EMULATED;
		return (0);
	case 4:
		strlcpy(aep->name, AudioEslinear_le, sizeof(aep->name));
		aep->encoding = AUDIO_ENCODING_SLINEAR_LE;
		aep->precision = 16;
		aep->flags = 0;
		return (0);
	case 5:
		strlcpy(aep->name, AudioEulinear_le, sizeof(aep->name));
		aep->encoding = AUDIO_ENCODING_ULINEAR_LE;
		aep->precision = 16;
		aep->flags = AUDIO_ENCODINGFLAG_EMULATED;
		return (0);
	case 6:
		strlcpy(aep->name, AudioEslinear_be, sizeof(aep->name));
		aep->encoding = AUDIO_ENCODING_SLINEAR_BE;
		aep->precision = 16;
		aep->flags = AUDIO_ENCODINGFLAG_EMULATED;
		return (0);
	case 7:
		strlcpy(aep->name, AudioEulinear_be, sizeof(aep->name));
		aep->encoding = AUDIO_ENCODING_ULINEAR_BE;
		aep->precision = 16;
		aep->flags = AUDIO_ENCODINGFLAG_EMULATED;
		return (0);
	default:
		return (EINVAL);
	}
}

int
vad_set_params(v, setmode, usemode, play, rec)
	void *v;
	int setmode, usemode;
	struct audio_params *play, *rec;
{
	struct vad_softc *sc = v;
	int error = 0;

	
	DPRINTF(("vad_setmode()\n"));
	
	/* Signal to svad that the encoding info has changed */
	svad_softcs[sc->sc_dev.dv_unit].sc_flags |= VAD_PARAM;
	
	if (setmode & AUMODE_PLAY) {
		play->factor = 1;
		play->sw_code = NULL;
		switch(play->encoding) {
		case AUDIO_ENCODING_ULAW:
			play->factor = 2;
			play->sw_code = mulaw_to_slinear16;
			break;
		case AUDIO_ENCODING_SLINEAR_LE:
			if (play->precision == 8)
				play->sw_code = change_sign8;
			break;
		case AUDIO_ENCODING_ULINEAR_LE:
			if (play->precision == 16)
				play->sw_code = change_sign16;
			break;
		case AUDIO_ENCODING_ALAW:
			play->factor = 2;
			play->sw_code = alaw_to_slinear16;
			break;
		case AUDIO_ENCODING_SLINEAR_BE:
			if (play->precision == 16)
				play->sw_code = swap_bytes;
			else
				play->sw_code = change_sign8;
			break;
		case AUDIO_ENCODING_ULINEAR_BE:
			if (play->precision == 16)
				play->sw_code = change_sign16_swap_bytes;
			break;
		default:
			return (EINVAL);
		}

		/* Give the drivers access to this encoding info */
		sc->sc_ec.sample_rate = play->sample_rate;
		sc->sc_ec.encoding = play->encoding;
		sc->sc_ec.precision = play->precision;
		sc->sc_ec.channels = play->channels;

		/* sample rate and channels will be the same */	
		svad_softcs[sc->sc_dev.dv_unit].sc_ec = sc->sc_ec;
		
		return (error);
	}

	if (setmode & AUMODE_RECORD) {
		rec->factor = 1;
		rec->sw_code = 0;
		switch(rec->encoding) {
		case AUDIO_ENCODING_ULAW:
			rec->sw_code = ulinear8_to_mulaw;
			break;
		case AUDIO_ENCODING_SLINEAR_LE:
			if (rec->precision == 8)
				rec->sw_code = change_sign8;
			break;
		case AUDIO_ENCODING_ULINEAR_LE:
			if (rec->precision == 16)
				rec->sw_code = change_sign16;
			break;
		case AUDIO_ENCODING_ALAW:
			rec->sw_code = ulinear8_to_alaw;
			break;
		case AUDIO_ENCODING_SLINEAR_BE:
			if (rec->precision == 16)
				rec->sw_code = swap_bytes;
			else
				rec->sw_code = change_sign8;
			break;
		case AUDIO_ENCODING_ULINEAR_BE:
			if (rec->precision == 16)
				rec->sw_code = swap_bytes_change_sign16;
			break;
		default:
			return (EINVAL);
		}

		return (error);
	}

	return (0);
}

int
vad_round_blocksize(v, blk)
	void *v;
	int blk;
{
	
	// struct vad_softc *sc = v;
	int blksize;
	
	DPRINTF(("vad_round_blocksize(blk=%d) return = %d\n", blk, blk & ~0x3f));

	blksize = 8192;

	return(blksize);

}

int
vad_halt_output(v)
	void *v;
{
	struct vad_softc *sc = v;
	int unit = sc->sc_dev.dv_unit;
	struct svad_softc *svad_sc = &svad_softcs[unit];
	
	DPRINTF(("%s: halt_output\n", sc->sc_dev.dv_xname));
	
	svad_sc->sc_flags   &= ~VAD_BUSY;

	return 0;
}

int
vad_halt_input(v)
	void *v;
{

#ifdef AUDIO_DEBUG	
	struct vad_softc *sc = v;
#endif
	
	DPRINTF(("%s: halt_input\n", sc->sc_dev.dv_xname));

	return 0;
}

int
vad_getdev(v, adp)
	void *v;
	struct audio_device *adp;
{
	struct vad_softc *sc = v;
	DPRINTF(("vad_getdev()\n"));
	*adp = sc->sc_audev;
	return 0;
}

int
vad_set_port(v, cp)
	void *v;
	mixer_ctrl_t *cp;
{
   /*
	struct vad_softc *sc = v;
   */
        
        DPRINTF(("vad_set_port: port=%d num_channels=%d\n",
                    cp->dev, cp->un.value.num_channels));
	return(0);
}

int
vad_get_port(v, cp)
	void *v;
	mixer_ctrl_t *cp;
{
   /*
	struct vad_softc *sc = v;
   */

        DPRINTF(("vad_get_port: port=%d num_channels=%d\n",
                    cp->dev, cp->un.value.num_channels));
	return(0);
}

int
vad_query_devinfo(v, dp)
	void *v;
	mixer_devinfo_t *dp;
{
    /*
	struct vad_softc *sc = v;
    */

        DPRINTF(("vad_query_devinfo: index=%d\n", dp->index));
	return(EINVAL);
}
void *
vad_allocm(addr, direction, size, pool, flags)
	void *addr;
	int direction;
	size_t size;
	int pool;
	int flags;

{
	// struct vad_softc *sc = addr;
	void * p;	

	/* Allocate the buffer for the MI audio driver */	
	DPRINTF(("vad_allocm(size=%d)\n",size));
	if((p = malloc(size, pool, flags)) == NULL)  { 
		DPRINTF(("vad_allocm: audio allocation failed\n"));
		return(NULL);
	}
	
	bzero(p, size);
	
	return p;
}

void
vad_freem(v, ptr, pool)
	void *v;
	void *ptr;
	int pool;
{
	// struct vad_softc *sc = v;

	DPRINTF(("vad_freem() success\n"));
	free(ptr, pool);
}


int
vad_get_props(v)
	void *v;
{
	/* return AUDIO_PROP_MMAP | AUDIO_PROP_INDEPENDENT | AUDIO_PROP_FULLDUPLEX; */
	DPRINTF(("vad_props()\n"));
	return AUDIO_PROP_INDEPENDENT | AUDIO_PROP_FULLDUPLEX; 
}

int
vad_trigger_output(v, start, end, blksize, intr, arg, param)
	void *v;
	void *start, *end;
	int blksize;
	void (*intr)(void *);
	void *arg;
	struct audio_params *param;
{
	struct vad_softc *sc = v;
	int unit = sc->sc_dev.dv_unit;
	struct svad_softc *svad_sc = &svad_softcs[unit];

	DPRINTF(("vad_trigger_output(unit = %d)\n", unit));
		
	/* 
	 * To signal to the svad device we a ready we just fill in the
	 * correct pointers so that it can access the audio ring buffer
	 * and mark the device as busy.
         */
	
	/* Now that we know the blocksize, allocate enough memory for our temp buffer */
	DPRINTF(("vad_trigger_output: allocate temp buffer\n"));
	if((svad_sc->sc_tbuf = malloc(blksize, M_DEVBUF, M_WAITOK)) == NULL)  { 
		DPRINTF(("vad_trigger_output: allocate temp buffer failed\n"));
		return(-1);
	}


	/* Give the drivers access to this encoding info */
	sc->sc_ec.sample_rate = param->sample_rate;
	sc->sc_ec.encoding = param->encoding;
	sc->sc_ec.precision = param->precision;
	sc->sc_ec.channels = param->channels;
	svad_sc->sc_ec = sc->sc_ec;

	svad_sc->sc_audsc = sc->sc_audsc;
	
	svad_sc->sc_blen     = 0;     /* This tells us how much is used */	
	svad_sc->sc_toff     = 0;
	svad_sc->sc_acbstart = start;
        svad_sc->sc_acbend   = end;
	svad_sc->sc_outp     = start;
	svad_sc->sc_blksize  = blksize;
	svad_sc->sc_intrp    = intr;
	svad_sc->sc_argp     = arg;
	svad_sc->sc_flags   |= VAD_BUSY;
	svad_sc->sc_flags   |= VAD_PARAM;

	/* Wakeup svad just in case it was blocking for us */
	wakeup(&svad_sc->sc_rchan);	
	
	DPRINTF(("vad_trigger_output(%x, %x, %d, %p, %p, %p)\n",
	    kvtop((caddr_t)start), kvtop((caddr_t)end),
	    blksize, intr, arg, param));

	
	return 0;

}

int
vad_trigger_input(v, start, end, blksize, intr, arg, param)
	void *v;
	void *start, *end;
	int blksize;
	void (*intr)(void *);
	void *arg;
	struct audio_params *param;
{
	// struct vad_softc *sc = v;

	DPRINTF(("vad_trigger_input(%x, %x, %d, %p, %p, %p)\n",
	    kvtop((caddr_t)start), kvtop((caddr_t)end),
	    blksize, intr, arg, param));

	return 0;
}




/* HERE IS THE VAD SLAVE ENTRY POINTS */

// svad function prototypes
int svadopen(dev_t, int, int, struct proc *);
int svadclose(dev_t, int, int, struct proc *);
int svadread(dev_t, struct uio *, int);
int svadwrite(dev_t, struct uio *, int);
int svadioctl(dev_t,  u_long, caddr_t, int, struct proc *);

	
/* XXX This does not get called by autoconf, it is called by vad_attach */
void
svadattach(n)
	int n;
{
	DPRINTF(("svadattach(n = %d)\n", n));
	
	if(n <= 0)
		return;

	if(svad_softcs == NULL)  {
	
		svad_softcs = malloc(n * sizeof(struct svad_softc), M_DEVBUF, M_WAITOK);
		if(svad_softcs == NULL)  {
			DPRINTF(("svadattach: malloc failed\n"));
			return;
		}
	
		bzero(svad_softcs, n * sizeof(struct svad_softc));

	}
	
}




int
svadopen(dev, flag, devtype, p)
	dev_t dev;
	int flag;
	int devtype;
	struct proc * p;

{
	
	int unit = minor(dev);
	struct svad_softc *sc = &svad_softcs[unit];

	DPRINTF(("svad_open(flags=%d)\n", flag));
	
	if(svad_softcs == NULL)  {
		DPRINTF(("svadopen: softcs not allocated\n"));
		return(ENXIO);
	}

	/* This svad has not  been opened then go ahead */
	if( (sc->sc_flags & SVAD_BUSY) == 0 )
	       sc->sc_flags |= SVAD_BUSY;
	/* Otherwise we say we are busy */	
	else
		return(EBUSY);	
	
	return(0);

}


int
svadclose(dev, flag, devtype, p)
        dev_t dev;
        int flag;
        int devtype;
        struct proc * p;
 
{

	int unit = minor(dev);
	struct svad_softc *sc = &svad_softcs[unit];
	
	DPRINTF(("svad_close(flags=%d)\n", flag));
 	
	/* Mark this svad as not busy */
	sc->sc_flags &= ~SVAD_BUSY;
        
	return(0);
 
}


int
svadread(dev, uio, flag)
        dev_t dev;
        struct uio *uio;
        int flag;

{
	
#define BUFSIZ		100	/* amount to uiomove at a time */
	
	int unit = minor(dev);
	struct svad_softc *sc = &svad_softcs[unit];
	int error = 0, cc, tcc=0, i;
	struct audio_params_pack pack;
	
	DPRINTF(("svad_read(unit = %d)\n", unit)); 
	
	/* We sleep if the trigger_output has not been called on that side */
	while( ((sc->sc_flags & VAD_BUSY) == 0) )  {

		if(flag&FNONBLOCK)
			return(EWOULDBLOCK);

		DPRINTF(("svadread: sleeping"));
		
		error = tsleep(&sc->sc_rchan, PWAIT | PCATCH, "svadread: sleep" , 0);
	
		if(error)
			return(error);
		
	}	

	
	
	
	/* If the audio encoding parameters of VAD have changed then we must */
	/* must let the user reading off svad know. To do this we fill the   */
	/* requesting buffer with a special byte pattern. This lets them     */
        /* know the audio params have changed. Make sure the packet wil fit  */
	if( ((sc->sc_flags & VAD_PARAM) != 0) && uio->uio_resid >= sizeof(struct audio_params_pack) )  {
	
		DPRINTF(("svadread: sending audio parameter packet\n"));
		
		sc->sc_flags &= ~VAD_PARAM;
		
		/* Now we set what the data is going to look like */
		/* as it comes out, after software decoding       */
	
		
		/* If it is mulaw, alaw, we will decode it to 16 bit signed linear */
		if(sc->sc_ec.encoding == AUDIO_ENCODING_ULAW ||	sc->sc_ec.encoding == AUDIO_ENCODING_ALAW) {
			sc->sc_ec.encoding = AUDIO_ENCODING_SLINEAR_LE;
			sc->sc_ec.precision = 16;
		}
	
		/* if it was big endian, it will be litte endian */	
		if(sc->sc_ec.encoding == AUDIO_ENCODING_SLINEAR_BE)
		       	sc->sc_ec.encoding = AUDIO_ENCODING_SLINEAR_LE;

		if(sc->sc_ec.encoding == AUDIO_ENCODING_ULINEAR_BE)
		       	sc->sc_ec.encoding = AUDIO_ENCODING_ULINEAR_LE;  

		/* Otherwise it will be the same */
		


		/* Set up the audio parameter packet */

		/* Fill the id up, note only 8 chars, no NULL terminator */
		for(i=0;i<8;i++) 
			pack.id[i] = byte_patt[i];	
	
		/* The size is the packet minus the id and this integer*/
		pack.size = sizeof(struct audio_params_pack) - 8 - sizeof(int);
		
		/* Copy the audio parameter in */
		pack.ec = sc->sc_ec;	

		/* Move the packet into the buffer */
		error = uiomove((caddr_t)&pack, sizeof(struct audio_params_pack), uio);

		if(error)
			return(error);
		
	}
	
	/* Now we want to block for a moment if there is not enough data on */
	/* the MI audio driver's buffer to read, we don't want any silence  */
	/* We don't want to block if they are draining the audio buffer too */
	while( sc->sc_audsc->sc_pr.used < (uio->uio_resid + (sc->sc_blksize * 2)) && 
	      sc->sc_audsc->sc_pr.usedlow != sc->sc_blksize )  {

		if(flag&FNONBLOCK)
			return(EWOULDBLOCK);

		DPRINTF(("svadread: sleeping for data"));

		/* We only sleep for a limited time, we can't wait for someone to wake us up */
		/* because nothing on the other side will be called again 		     */
		error = tsleep(&sc->sc_rchan, PWAIT | PCATCH, "svadread: sleep", hz / 32);
		
		if(error && error != EWOULDBLOCK)
			return(error);	
		
	}
			
	/* If there is any data in our temp buffer then transfer that first */
	while(uio->uio_resid > 0 && sc->sc_blen > 0)  {
		cc = min(uio->uio_resid, BUFSIZ);
		cc = min(cc, sc->sc_blen);
		error = uiomove(sc->sc_tbuf+sc->sc_toff, cc, uio);

		if(error)  {
			DPRINTF(("svadread: tbuf uiomove failed\n"));
			return(error);
		}

		sc->sc_blen-=cc;	
		sc->sc_toff+=cc;

		if(sc->sc_blen == 0)
			sc->sc_toff = 0;
		
	}
	

	/* XXX The code below is really messy but it works, I want to clean it up later */
	
	/* Transfer all the complete blocks */
	while(uio->uio_resid >= sc->sc_blksize)  {

		/* Transfer one block */
		while(tcc < sc->sc_blksize) {
			cc = min(uio->uio_resid, BUFSIZ); 
		        cc = min(cc, (sc->sc_blksize - tcc));
			error = uiomove(sc->sc_outp, cc, uio);

			if(error)  {
				DPRINTF(("svadread: uiomove failed\n"));
				return(error);
			}

			tcc+=cc;	
			sc->sc_outp+=cc;	
		}
		
		tcc = 0;
	
		if(sc->sc_outp >= sc->sc_acbend)
			sc->sc_outp = sc->sc_acbstart;
		
		/* Call the interrupt routine to show that we transferred one block */
		sc->sc_intrp(sc->sc_argp);
	
		/* Make sure we are still open */
		if((sc->sc_flags & VAD_BUSY) == 0) 
			return(0);
		
	}

	/* Else we need to transfer one whole block and copy the change into a temp buffer */
	while(uio->uio_resid < sc->sc_blksize && uio->uio_resid > 0)  {	

		/* Transfer the remaining partial block */
		while(uio->uio_resid > 0)   {
			cc = min(uio->uio_resid, BUFSIZ);
			error = uiomove(sc->sc_outp, cc, uio);

			if(error)  {
				DPRINTF(("svadread: uiomove failed\n"));
				return(error);
			}

			tcc+=cc;
		        sc->sc_outp+=cc;
		}		

		/* Copy whatever is left over to a temp buffer for the next read */
		sc->sc_blen = sc->sc_blksize - tcc;
		memcpy(sc->sc_tbuf, sc->sc_outp, sc->sc_blen);
		
		sc->sc_outp+=sc->sc_blen;

		if(sc->sc_outp >= sc->sc_acbend)
			sc->sc_outp = sc->sc_acbstart;
		
		/* Call the interrupt routine to show that we transferred one block */
		sc->sc_intrp(sc->sc_argp);

		/* Make sure we are still open */
		if((sc->sc_flags & VAD_BUSY) == 0) 
			return(0);
		
	}	
	
	return(0);

}


int
svadioctl(dev, cmd, data, flag, p)
        dev_t dev; 
        u_long cmd;
        caddr_t data;
        int flag;
        struct proc *p;
{

	int unit = minor(dev);
	struct svad_softc *sc = &svad_softcs[unit];
	struct svad_info *info = (struct svad_info *)data;
	
	DPRINTF(("\nsvadioctl is called\n"));
	
	switch(cmd) {
		case SVAD_GETINFO:
			info->ei = sc->sc_ec;	
		break;
		/* SVAD_GETSTATUS returns 0 if vad_trigger_output hasn't */
		/* been called and 1 otherwise. 			 */
		case SVAD_GETSTATUS:
			*(int *)data = ((sc->sc_flags & VAD_BUSY) != 0);
		break;
	}
		
	return(0);

}
