#include <stdio.h>
#include <fcntl.h>
#include <errno.h>

#if defined (OPENBSD_VAD)
#include <soundcard.h>
#endif

#if defined (LINUX_MAUDIO)
#include <sys/soundcard.h>
#endif

#include <vorbis/vorbisenc.h>

#include "vadclient.h"
#include "audio.h"


struct audio_device aud_dev;		/* Our audio device structure         */
int playback_started=0;			/* Has playback started?              */
int lowat=0;				/* low watermark for playing data   */
int flush_audio_buffer=0;		/* Flag, flush audio buffers?	    */

#define BYTES_PER_SEC   (aud_dev.dev_params.channels    * \
			aud_dev.dev_params.sample_rate * \
			(aud_dev.dev_params.precision / 8))


/* Open the audio device, O_NDELAY for non-blocking */
int audio_open() { return (aud_dev.fd_dsp = open("/dev/dsp", O_WRONLY | O_NDELAY)); }
	
/* close the audio device */
void audio_close() { close(aud_dev.fd_dsp); }
	



/* Closes and reopens the audio device, then sets parameters */
void audio_set_params(struct audio_params *ap)
{

	int channels, format, rate;

	/* Reset the device only if playback has started   */
	/* We do this because the OSS Audio guide says its */
	/* bad to reset the audio device after it was just */
	/* opened for playback.				   */
	if(playback_started != 0) {

		/* Reset the audio device, stop playback */
		if(ioctl(aud_dev.fd_dsp, SNDCTL_DSP_RESET, NULL) < 0) 
			error_exit("audio_set_params: reset failed\n");

		/* Close the device */
		audio_close();

		/* Reopen it */
		if( audio_open() == -1)  
			error_exit("REopen /dev/dsp");
		

	}
	
	/* If ioctls return an error, than we die    */
	/* But just because we didn't get an error   */
	/* doesn't mean that the parameters were set */
	/* correctly. We must check		     */

	/* Set either stereo or mono mode */
	channels = ap->channels;

	/* Set it */	
	if(ioctl(aud_dev.fd_dsp, SNDCTL_DSP_CHANNELS, &channels) == -1)  
		error_exit("audio_set_params: SNDCTL_DSP_CHANNELS");

	/* Couldn't set audio channels to ap->channels */
	if(channels != ap->channels) 
		printf("Error: Couldn't configure %d channel(s)!\n", ap->channels);

	aud_dev.dev_params.channels = channels;
	printf("Configured %d channel(s) for playback\n", channels);
	
	
	

	format = ap->precision;
	
	if(ioctl(aud_dev.fd_dsp, SNDCTL_DSP_SETFMT, &format) == -1)  
		error_exit("audio_set_params: SNDCTL_DSP_SETFMT");

	/* 8-bit and 16-bit PCM are unsupported !!! */
	if(format != AFMT_S16_LE && format != AFMT_U8)
		error_exit("audio_set_params: 8 and 16 bit PCM unsupported !!!");

	if(format != ap->precision)
		printf("Error: %d-bit precision is not supported\n", ap->precision);

	aud_dev.dev_params.precision = format; 
	printf("Configured %d-bit precision for playback\n", format);
	

	

	rate = ap->sample_rate;
	
	if(ioctl(aud_dev.fd_dsp, SNDCTL_DSP_SPEED, &rate) == -1)  
		error_exit("audio_set_params: SNDCTL_DSP_SPEED");
	
	if(rate != ap->sample_rate)
		printf("Error: %d Hz sample rate not supported\n", ap->sample_rate);
		
	aud_dev.dev_params.sample_rate = rate;
	printf("Configured %d Hz sample rate for playback\n", rate);



	
	/* Now find out the block size */
	if( ioctl(aud_dev.fd_dsp, SNDCTL_DSP_GETBLKSIZE, &aud_dev.blocksize) == -1)
		error_exit("audio_set_params: SNDCTL_DSP_GETBLKSIZE");

	printf("Audio Device blocksize is %d\n\n", aud_dev.blocksize);
	
	/* Now that we have the format data we can calculate a low */
	/* watermark. This is just so we can buffer a little data  */
	/* at the start of streaming to absorb a network hiccups.  */
	/* This is an arbitrary length of audio.		   */
	lowat = (aud_dev.dev_params.precision / 8) * 
		aud_dev.dev_params.sample_rate     * 
		aud_dev.dev_params.channels        *
		SECS_TO_BUFFER;

	aud_dev.set_flag = 1;
	playback_started = 0;	
	flush_audio_buffer=0;	

}




/* Write the buffered audio data to our device, return negative if failed */
int write_to_audio(struct ringbuffer *devbuf)
{
	int nw, space, total_written=0;

	if(devbuf == NULL)
		return(-1);

	if ( (server_time-1) < audio_time && !flush_audio_buffer)
		return(0);
	
	
	/* We write to audio device while we still have data left to write */
	/* and the device is set up for us.				   */
	while(devbuf->used > 0 && aud_dev.set_flag == 1) {

		/* Signal that we have written data to the audio device */
		/* This is a simple flag that allows us to see if we    */
		/* have started playing audio data yet.		        */
		playback_started = 1;
		
		/* How much space is left till the end pointer */
		space = devbuf->end - devbuf->outp;
		
		/* write everything we can in one go */
		if(space >= devbuf->used)
			nw = write(aud_dev.fd_dsp, devbuf->outp, devbuf->used);
		/* Otherwise, we write just till the end */
		else 
			nw = write(aud_dev.fd_dsp, devbuf->outp, space);
		
		/* If we have any error, other than EAGAIN, exit */
		if(nw < 0 && errno != EAGAIN)
			return(-1);

		/* If we wrote some data to the device */
		else if(nw > 0)  {
				
			devbuf->outp+=nw;
			total_written+=nw;

			if(devbuf->outp >= devbuf->end)
				devbuf->outp = devbuf->start;
				
			devbuf->used-=nw;
		
			/* Add the amount of data we played to the audio time */
			/* ie. if we played 5 seconds of audio data, and 5    */
			/* seconds to the audio time.			      */
			audio_time+=((double)nw/(double)BYTES_PER_SEC)*1000000.0; 
		}
			
		/* We recieved EAGAIN, the audio device is full */
		else if(errno == EAGAIN) 
			break;
			
	}
	
	/* Success ... in some shape or form */
	return(total_written);
	
}


void set_buffer_flush() { flush_audio_buffer = 1; }
void unset_buffer_flush() { flush_audio_buffer = 0; }


int get_precision()  { return aud_dev.dev_params.precision;    }
int get_samplerate() { return aud_dev.dev_params.sample_rate;  }
int get_channels()   { return aud_dev.dev_params.channels;     }
int get_blocksize()  { return aud_dev.blocksize; 	       }
int is_audio_set()   { return aud_dev.set_flag;		       }


int audio_frags() 

{

	struct audio_buf_info info;

	ioctl(aud_dev.fd_dsp, SNDCTL_DSP_GETOSPACE, &info);
	
	return(info.fragments);

}

int audio_fragstotal() 

{

	struct audio_buf_info info;

	ioctl(aud_dev.fd_dsp, SNDCTL_DSP_GETOSPACE, &info);
	
	return(info.fragstotal);

}
