#include <stdio.h>
#include <vorbis/vorbisenc.h>
#include "vadclient.h"
#include "compression.h"

/*
 * a hexdump routine
 */

#include <stdarg.h>
#include <string.h>

/* All of the ogg vorbis state information variables */

ogg_sync_state   oy; 			/* sync and verify incoming physical bitstream */
ogg_stream_state os; 			/* take physical pages, weld into a logical
			  			stream of packets */
ogg_page         og; 			/* one Ogg bitstream page.  Vorbis packets are inside */
ogg_packet       op; 			/* one raw packet of data for decode */
  
vorbis_info      vi; 			/* struct that stores all the static vorbis bitstream settings */

vorbis_comment   vc; 			/* struct that stores all the bitstream user comments */
vorbis_dsp_state vd; 			/* central working state for the packet->PCM decoder */
vorbis_block     vb; 			/* local working space for packet->PCM decode */

ogg_int16_t convbuffer[8192]; 		/* take 8k out of the data segment, not the stack */
int convsize=8192;

int vorbis_setup=0;  			/* marks whether we have setup vorbis before */

#define BYTES_PER_SEC   (vi.channels*2*vi.rate)

/* Initializes the encoding engine. Must pass it a correctly 		*/
/* reconstructed net_audio_params_pack with ogg_packets for headers	*/
/* Return 0 if sucessfull 						*/
int decode_init(struct net_audio_params_pack *packet)
{

	/* This allows us to begin reading pages */
	ogg_sync_init(&oy);

	/* We have the header packets already so this should be easy */
	
	/* Pass in the serialno of the stream */
	ogg_stream_init(&os, packet->serialno);

	vorbis_info_init(&vi);
	vorbis_comment_init(&vc);

	/* We just pass the headers in */
	if( vorbis_synthesis_headerin(&vi, &vc, &packet->headers[0]) < 0 ||
	    vorbis_synthesis_headerin(&vi, &vc, &packet->headers[1]) < 0 ||		
	    vorbis_synthesis_headerin(&vi, &vc, &packet->headers[2]) < 0  )  {

		/* Uh Oh! These headers are corrupt! Return error */
		printf("decode_init: error - header corruption\n");
		return(-1);

	}
	
	/* Set up the packet->PCM decoder */
	vorbis_synthesis_init(&vd, &vi);
	vorbis_block_init(&vd, &vb);

    	convsize=4096/vi.channels;

	/* Mark vorbis as being setup */
	vorbis_setup = 1;
	
	/* Success */
	return(0);
	
}


/* Simply cleans up */
void decode_deinit()
{

	/* Clear all of the data structures and state variables for the engine */
	ogg_stream_clear(&os);
	vorbis_block_clear(&vb);
	vorbis_dsp_clear(&vd);
	vorbis_comment_clear(&vc);
	vorbis_info_clear(&vi);
	ogg_sync_clear(&oy);
	
}

/* Decodes a buffer of ogg stream data */
int decode_and_write(struct audio_data_packet *packet)

{

	char *rbuffer;		    /* Pointer for packet->data 	 */
	int rlen;		    /* Length of the packet->data        */
	unsigned long long tstamp;  /* Timestamp of the data             */
	char *buffer;		    /* Buffer to submit data             */
	int result;		    /* Result from ogg library functions */		
	static int print_once = 0;
	
	/* if the packet pointer is null 	    */
	/* or there is no audio data in this packet */
	if(packet == NULL || packet->hdr.len == 0)  
		return(0);
	
	/* If the pointer and length of the packet data legit the we  */
	/* can get the data pointer and the length of the data.	      */ 
	rbuffer = packet->data;	
	rlen = packet->hdr.len;
	tstamp = packet->hdr.timestamp;
		
	/* Submit our data to the engine */
	buffer=ogg_sync_buffer(&oy,rlen);
	memcpy(buffer, rbuffer, rlen);
	ogg_sync_wrote(&oy,rlen);

	result = ogg_sync_pageout(&oy, &og);
	
	/* If we dont have a page, just return the number of bytes we read in */
	if(result <= 0) 
		return(rlen);
	
	
	/* Submit our page */
	ogg_stream_pagein(&os, &og);
	

	/* Get any packets we have */
	while(1) {
			
		result=ogg_stream_packetout(&os, &op);

		/* If there are no packets left */	
		if(result == 0) 
			break;
		

		/* we have a packet */
		else {
				
			float **pcm;
			int samples;

			if(vorbis_synthesis(&vb,&op)==0)
				vorbis_synthesis_blockin(&vd,&vb);
							
			while((samples=vorbis_synthesis_pcmout(&vd,&pcm))>0) {
				int i, j, clipflag=0;
				int bout=(samples<convsize?samples:convsize);
				int audio_data_size = bout*2*vi.channels;				  
							   
				for(i=0;i<vi.channels;i++){
		  			ogg_int16_t *ptr=convbuffer+i;
					float  *mono=pcm[i];
		  					      
					for(j=0;j<bout;j++){
		    				int val=mono[j]*32767.f;
		   
		    				if(val>32767){
		      					val=32767;
		      					clipflag=1;
			    			}
			    			if(val<-32768){
			      				val=-32768;
			      				clipflag=1;
		    				}
								
						*ptr=val;
		    				ptr+=vi.channels;
		  						    
					}
							   
				}
		
				/* convert from little endian to whatever we need */
				for(j=0;j<audio_data_size/2;j++) 
					convbuffer[j] = LETOH16(convbuffer[j]);
				
				/* write to audio buffer */			
				write_to_buffer((char *)convbuffer,
				audio_data_size, tstamp);

				/* increase the timestamp by audio_data_size  */
				/* milliseconds worth of time.		      */
				tstamp+= ((double)audio_data_size / 
					  (double)BYTES_PER_SEC) * 
					 1000000.0;
				
				/* Tell the engine how much we wrote */
				vorbis_synthesis_read(&vd, bout);	
			
			}
					
		}
						
					
	}

	return(rlen);

}


int is_decode_setup() { return vorbis_setup; }
