#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <time.h>
#include <netinet/in.h>
#include <rpc/rpc.h>     /* XDR */

#include <math.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <getopt.h>

#include <vorbis/vorbisenc.h>

#include "vadclient.h"
#include "compression.h"
#include "audio.h"




#define GROUP		"224.1.1.39"	/* Multicast Group to join */

#define RSIZE		8192	
#define BUFSIZE		532		/* # of blocks in our buffer */		
#define PORT		60123


/* Getopts stuff */
#define OPTION_STRING "p:m:r:"

static struct option long_options[] =
{
	{ "precision", required_argument, NULL, 'p' },
	{ "mode",      required_argument, NULL, 'm' },
	{ "rate",      required_argument, NULL, 'r' },
	{ 0, 0, 0, 0 }
};


/* PROTOTYPES */

/* print information */
static void print_info();

/* parse command line options */
static int parse_command_line(int argc, char *argv[]);

/* read the audio data off the network and into our ringbuffer */
static int read_from_network();



/* Conversion Routines */
static void swap_bytes(char *buffer, int blocksize);
static void mts(u_char *p, int cc);
static void linear8_to_linear16_le(u_char *p, int cc);
static void linear16_to_linear8_le(u_char *p, int cc);


/* GLOBALS */
struct audio_params params;		/* Recieved audio parameters        */
struct ringbuffer devbuf;		/* Audio Buffer to write to device  */

struct net_audio_params_pack pack;	/* The audio paramter packet 	    */
int param_packet_recieved=0;		/* Did we recieve a paramter pack   */

int fd_sock;				/* Socket descriptor 		    */

int total_packets=0;			/* total packets recieved           */
int total_bytes=0;			/* total bytes transfered 	    */

int bit_precision=0;			/* set by command, force precision  */
int sample_rate=0;			/* set by command, force rate       */
int num_channels=0;			/* set by command, force # channels */



unsigned long long server_time = -1;	/* -1 marks initialize 		    */	
unsigned long long audio_time = -2;	/* -2 marks initialize 		    */		


unsigned char header_data[3][RSIZE*4];  /* store header, ogg_packet's */
					/* we want stack space.	      */




/* Simple Max Function */
unsigned long long max(unsigned long long a1, unsigned long long a2)
{
	if(a1 >= a2)
		return a1;
	else
		return a2;		
}


int main(int argc, char *argv[])

{
	
	struct sockaddr_in loc_addr;		/* Local address        */
	int loc_addrlen = sizeof(loc_addr);	/* local address length */
	struct ip_mreq mreq;			/* Multicast request    */
	
	printf("VAD - Virtual Audio Driver - Client Player\n\n");
	
	/* parse command line options, if return is negative we */
	/* have an error. Exit the program.			*/
	if(parse_command_line(argc, argv) < 0)
		return(-1);


	

	
	/* Open the audio device */
	if(audio_open() == -1)  {
		perror("open /dev/dsp");
		return(errno);
	}



	
	/* Open a Datagram Socket */
	if((fd_sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		perror("open socket failed"); 
		audio_close();
		return(errno);
	}

	/* Setup the local address for the socket */
	bzero(&loc_addr, sizeof(loc_addr));
	loc_addr.sin_family = AF_INET;
	loc_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	loc_addr.sin_port = htons(PORT);

	if(bind(fd_sock, (struct sockaddr *)&loc_addr, loc_addrlen) < 0) 
		error_exit("socket bind failed\n");
					 
	/* Join the multicast group  */
	mreq.imr_multiaddr.s_addr = inet_addr(GROUP);
	mreq.imr_interface.s_addr = htonl(INADDR_ANY);
	if (setsockopt(fd_sock, IPPROTO_IP, IP_ADD_MEMBERSHIP,
	    &mreq, sizeof(mreq)) < 0) {
		
		perror("setsockopt mreq");
		exit(1);
	}
				      
	
	

	/* Loop Forever, or until we break from within */
	while(1)  { 

		
		if(read_from_network() < 0)
			error_exit("main: read from network");
		
		
		if(param_packet_recieved == 1) {
		
			int ret=0;
		
			if(params.ogg_stream == 1) {
				/* Setup Ogg */

				/* If decode_init has been called before, call decode_deinit */
				if(is_decode_setup())  
					decode_deinit();
		
				ret = decode_init(&pack);
			}
			

			/* If we get an error the headers must be corrupt */
			if(ret != 0) {
				printf("Decode Init Failed\n");
				return(-1);
			}
		

			
			/* If the user specified it on the command */
			/* line then we need to ignore the audio   */
			/* paramters and force the device into the */
			/* state that  they ask. We don't care if  */
			/* the audio will sound like crap.         */	
			if(bit_precision != 0 ||
			   sample_rate != 0   ||
		   	   num_channels != 0  	) {
     				
				/* we have to make a temporary params struct */   
				/* so that we maintain the real audio params */
				/* Remember, we are just forcing the device  */
				/* into a certain format, not the audio data */
				/* itself. We will try our best to convert.  */
				struct audio_params t_params;

				/* Make a copy of the audio  params */
				memcpy(&t_params, &params, sizeof(struct audio_params));	
				
				/* Now just change what the user wants to   */
				if(bit_precision != 0)
					t_params.precision = bit_precision;	
				if(sample_rate != 0)
					t_params.sample_rate = sample_rate;	
				if(num_channels != 0)
					t_params.channels = num_channels;	
			
				/* Now just set the forced parameters */
				audio_set_params(&t_params);	
				
			}	
				
			else /* use the params of actual audio data */
				audio_set_params(&params);
	



			/* audio_set_params gets the blocksize, now we allocate */
			/* our ringbuffer again based on multiples of this      */
			/* blocksize.						*/
			
			/* If we have allocated memory, free it */
			if(devbuf.start != NULL) {
				free(devbuf.start);
				devbuf.start = NULL;
			}
				
			devbuf.start = (unsigned char*)malloc(get_blocksize() * BUFSIZE);
			if(devbuf.start == NULL)
				error_exit("main: malloc audio buffer failed\n");
					
			devbuf.size = get_blocksize() * BUFSIZE;
			devbuf.inp = devbuf.outp = devbuf.start;
			devbuf.end = devbuf.start + devbuf.size;
			devbuf.used = 0;
	
			/* We handled the packet, so remove the flag */		
			param_packet_recieved=0;
	
			/* set total_packets for this stream back to zero */
			total_packets=0;
			
			/* We will need to record a new audio time */
			audio_time = -2;
			
		}
	

	
		/* print useful information */
		print_info();

		/* Write data to audio device, exit if failed */
		if(write_to_audio(&devbuf) < 0)	
			error_exit("main: write on audio");

	}

	
	
	/* Close the socket */
	close(fd_sock);
	
	/* Close the audio device */
	audio_close();	
	
	return(0);

}


/* parse any command line options, return 0 for success, negative otherwise */
int parse_command_line(int argc, char* argv[])
{
	char ch;
	
	/* Parse command line options */
	while((ch = getopt_long(argc, argv, OPTION_STRING, long_options, NULL)) != -1)  {
		switch(ch) {
			
			case 'p':  /* Force a precision */
				bit_precision = strtol(optarg, NULL, 10);
				
				if(bit_precision != 16 && bit_precision != 8) {
					printf("Error: invalid precision\n");
					return(-1);
				}
				
				break;
			
			case 'r':  /* Force sample rate */
				sample_rate = strtol(optarg, NULL, 10);
				break;
		
			case 'm':  /* Force number of channels */
				num_channels = strtol(optarg, NULL, 10);
				if(num_channels != 1 && num_channels != 2)  {
					printf("Error: invalid mode\n");
					return(-1);
				}

				break;
			
			case '?':
				return(-1);
				break;
		}
	
	}

	/* Success */
	return(0);
	
}




/* Prints useful info to the console like transfer rate and buffer usage */
void print_info()
{

	struct timeval tv;
	static int oTotal;		/* stores the old total bytes  */	
	double txRate=0.0;		/* Transfer rate	       */
	static int cur_time=0;		/* updated with gettimeofday() */

	/* Print out some information if the audio paramters have been set */
	if(is_audio_set() == 1)  {		

		gettimeofday(&tv, NULL);
			
		/* If time hasn't been set yet */
		if(cur_time == 0)
			cur_time = tv.tv_sec;
			
		if( (tv.tv_sec - cur_time) >= 1 ) {
		
			txRate = ((double)(total_bytes-oTotal)/(double)(tv.tv_sec-cur_time))/1024.0;
		
			oTotal = total_bytes;
			cur_time = tv.tv_sec;
			
			/* If we have some kind of active stream, print the stats */
			if(txRate > 0 || devbuf.used > 0) {	
				printf("packets = %d ; used = %d ; txRate = %.2f KB/s "
				       "; %d/%d frags free ; st = %lld ; at = %lld   \n",
				total_packets, devbuf.used, txRate, audio_frags(), 
				audio_fragstotal(), server_time, audio_time );
			}
				
		}			
				
			
	}


}



/* Close everything thing up and die when we get an error */
void error_exit(char *message)
{

	/* If we have allocated memory, free it */
	if(devbuf.start != NULL)
		free(devbuf.start);
	
	/* Print error if a system call hasn't failed*/
	if(errno == 0)
		printf(message);
	/* Otherwise, we use perror */
	else
		perror(message);	
	
	close(fd_sock);
	
	/* Close the audio device, assume it is open */
	audio_close();

	exit(errno);
	
}





/* read data off of the network into our buffer, return negative on error */
int read_from_network()
{
	
	struct sockaddr_in serv_addr;		/* Address of server          */
	int serv_addrlen = sizeof(serv_addr);   /* length of server address   */
	static int rbuflen=0;			/* length of data in rbuf     */
	static unsigned char rbuf[RSIZE*4];     /* recieve buffer	      */
	struct timeval pa_time;			/* packet arrival time        */
	
	/* This loop reads data off the network and into the audio buffer */
	/* If the buffer fills up the oldest record will be removed 	  */
	/* The loop will break from within if the socket starts to block  */
	
	
	while(1) {
			
		/* If there is no data left in rbuf     */
		/* Grab another packet off the network  */
		if(rbuflen == 0) {
			rbuflen = recvfrom(fd_sock, rbuf, RSIZE, MSG_DONTWAIT, 
			(struct sockaddr *)&serv_addr, &serv_addrlen); 
		}

		/* Error Check, anything but EAGAIN */	
		if(rbuflen < 0 && errno != EAGAIN) 
			error_exit("main: recvfrom");

		/* The socket is non-blocking, we dont want to wait if there */
		/* is no data to read in because there might be some left to */
		/* play. 						     */
		if(rbuflen == -1 && errno == EAGAIN)  {
			rbuflen = 0;
			break;
		}


		/* If it makes it to here, we have a packet to read */
		total_packets++;
				
	
		/* Add to the total bytes transefered */
		total_bytes+=rbuflen;
		

		/* Check if the first 8 bytes spell out the signature for an */
		/* audio parameter packet.   				     */
		if(rbuf[0] == 'N' && rbuf[1] == 'E' && rbuf[2] == 'T' &&
		   rbuf[3] == 'P' && rbuf[4] == 'A' && rbuf[5] == 'R' &&
		   rbuf[6] == 'A' && rbuf[7] == 'M')  {
	
			/* Used to decode XDR encoded audio parameter packet */
			XDR *dexdrs;		  	
			int i;
		
			dexdrs = (XDR *)malloc(sizeof(XDR));
			xdrmem_create(dexdrs, rbuf, RSIZE*4, XDR_DECODE);

		        /* Before we decode the packet, make sure we have */
			/* space to store the header packets if this is   */
			/* we have and ogg stream.			  */	
			pack.headers[0].packet = &header_data[0][0];
			pack.headers[1].packet = &header_data[1][0];
			pack.headers[2].packet = &header_data[2][0];
			
			/* Decode the packet */
			if(!xdr_net_audio_params_pack(dexdrs, &pack)) {        
				printf("XDR Decoding failers\n");
				return -1;
			}
		
			/* remove the packet data from rbuf */	
			memmove(rbuf, rbuf+xdr_getpos(dexdrs),
				rbuflen-xdr_getpos(dexdrs));
			rbuflen-=xdr_getpos(dexdrs);
			
			/* Destroy XDR decoding stream */
			xdr_destroy(dexdrs);	
			free(dexdrs);	
	

			/* first things first, update the server clock */
			server_time = max(pack.timestamp*1000000, server_time);
			
			/* If have reached the end of the stream than this is */
			/* probably the last control packet we will get. So   */
			/* we set a flag that tells the writing function to   */
			/* ignore the timestamps for synchronization. This    */
			/* allows us to finish playing the stream even though */
			/* we may have not gotten a control packet with a     */
			/* a late enough timestamp.			      */
			if(pack.end_of_stream)
				set_buffer_flush();
			
			
			/* Mark whether this is an ogg stream or not */
			params.ogg_stream=pack.ogg_flag;
			
				
			
			if(params.sample_rate != pack.sample_rate || 
			   params.precision != pack.precision     || 
			   params.channels != pack.channels 	  ||
			   params.ogg_stream_serial != pack.serialno)  {
			
				/* If a new audio stream, reset server_time */
				server_time = pack.timestamp;	
				
				/* Get the serial no */
				params.ogg_stream_serial = pack.serialno;
				
				/* Get all of the encoding paramters */	
				params.sample_rate = pack.sample_rate;
				params.precision = pack.precision;
				params.channels = pack.channels;
		
				printf("\nRecieved audio parameter packet\n");
				param_packet_recieved=1;
						
					
		
				/* Break from the loop if we have to set params */
				/* We leave the left over data in rbuf for next */
				/* time around 					*/
				break;
			}
	 				
			break;
						
		}


		/* If we recieved data that wasnt and audio param pack we try */
		/* to buffer it.					      */
		if(rbuflen > 0 && is_audio_set() == 1) {

			XDR *dexdrs;	
			struct audio_data_packet audio_packet;
			
			/* Create XDR stream to decode audio packet */	
			dexdrs = (XDR *)malloc(sizeof(XDR));
			xdrmem_create(dexdrs, rbuf, RSIZE*4, XDR_DECODE);

			/* If we failed to read in and audio_packet, ignore it */
			if(!xdr_audio_data_packet(dexdrs, &audio_packet)) {
				printf("Error: Invalid Audio Data Packet\n");
				rbuflen=0;
				break;
			}
		
			/* Destroy this xdr stream */
			xdr_destroy(dexdrs);
			free(dexdrs);

			
			/* If the audio_time is set to -2 that means we       */
			/* need to get it from the newest audio packet.       */
			if(audio_time == -2)	
				audio_time = audio_packet.hdr.timestamp;	
			
			/* If this is an ogg stream then we need to do some decoding first */
			if(params.ogg_stream == 1) {
				decode_and_write(&audio_packet);
				rbuflen=0;	
			}
		
			/* Otherwise we can just write it to the ringbuffer */	
			if(params.ogg_stream != 1) {  
				write_to_buffer(audio_packet.data, audio_packet.hdr.len, 
						audio_packet.hdr.timestamp);
				rbuflen=0;
			}
		
		}
	


		/* If parameters haven't been set yet and we get a packet */
		/* ignore it because we dont know how to play it.	  */
		if(rbuflen != 0 && is_audio_set() == 0)
			rbuflen = 0;
	
	}

	
	return(0);
	
}
	
		




/* Writes our data to the ring buffer */
int write_to_buffer(char *buffer, int size, unsigned long long timestamp)
{

	int off=0;

	/* These are conversions, if nescessary */
	if(params.ogg_stream != 1)  {
	
		if(params.precision == 8 && get_precision() == 16)  {
			linear8_to_linear16_le(buffer, size);
			size*=2;
		}
	
		if(params.precision == 16 && get_precision() == 8)  {
			linear16_to_linear8_le(buffer, size);
			size/=2;
		}
		
		if(params.channels == 1 && get_channels() == 2)  {
			mts(buffer, size);
			size*=2;
		}				  
	
	}
		
	/* Put the data into devbuf one byte at a time */
	while(size > 0)  {
				
		/* If we run out of space, erase oldest record */
		if(devbuf.used >= devbuf.size) {
			devbuf.outp++;
			devbuf.used--;
								
			if(devbuf.outp >= devbuf.end)
				devbuf.outp = devbuf.start;
		
		}
							
		*devbuf.inp = buffer[off];
		off++;
		devbuf.inp++;
		devbuf.used++;
	
		if(devbuf.inp >= devbuf.end)
			devbuf.inp = devbuf.start;
	
		size--;
						
	}

	/* Now that we have added data to the buffer we can attempt to write */
	/* some of it to the audio device.				     */
	if(write_to_audio(&devbuf) < 0)
		error_exit("write_to_buffer: write_to_audio");
	
}





/* Swap byte ordering, taken from tplay source */
static void swap_bytes(char *buffer, int blocksize)
{
	  register int i;
	  char c, *p;
	  p = buffer;
	  
	  for (i=0; i<blocksize/2; i++) {
	  	c = *p;
	  	*p = *(p+1);
	  	*++p = c;
	  	p++;
	  }
}

/* Mono to Stereo Conversion, taken from OpenBSD auconv.c */
void mts(u_char *p, int cc)
{
	u_char *q = p;
	
	p += cc;
	q += cc * 2;
	while ((cc -= 2) >= 0) {
		q -= 4;
		q[1] = q[3] = *--p;
		q[0] = q[2] = *--p;
	}

}

/* Convert signed linear 8 bit to 16 bit, from OpenBSD */
void linear8_to_linear16_le(u_char *p, int cc)
{
	u_char *q = p;

	p += cc;
	q += cc * 2;
	while (--cc >= 0) {
		q -= 2;
		q[1] = *--p;
		q[0] = 0;
	}
}


void
linear16_to_linear8_le(u_char *p, int cc)
{
	u_char *q = p;

	while ((cc -= 2) >= 0) {
		*q++ = p[1];
		p += 2;
	}
}

