#ifndef __VADCLIENT_H__
#define __VADCLIENT_H__

#define MAX_BUFSIZE      8192

#include <rpc/rpc.h> 	/* XDR */

/* XDR routine for long long's. */
/* xdr_longlong and xdr_hyper are defined in some versions of XDR. */
/* This routine is pretty much a copy from those routines.         */
static int xdr_long_long(xdrs, llp)
	XDR *xdrs;
	long long *llp;
{
	long l1, l2;

	/* If we are encoding */
	if(xdrs->x_op == XDR_ENCODE) {
		l1 = (long)((*llp) >> 32);
		l2 = (long)(*llp);
		return(XDR_PUTLONG(xdrs, &l1) && XDR_PUTLONG(xdrs, &l2));
	}

	/* If we are decoding */
	else if(xdrs->x_op == XDR_DECODE) {
		if(XDR_GETLONG(xdrs, &l1) && XDR_GETLONG(xdrs, &l2)) {
			*llp = ((long long) l1) << 32;
			*llp |= l2;
			return(1);
		}
		else
			return(1);
	}

	else if(xdrs->x_op == XDR_FREE)
		return(1);

	else
		return(0);

}

struct audio_params  {

	unsigned long  sample_rate;   
	unsigned int   encoding;      	   /* ulaw, alaw, based on BSD #define's */
	unsigned int   precision;     	   /* 8 bit or 16 bit  */
	unsigned int   channels;      	   /* mono=1, stereo=2 */
	unsigned int   ogg_stream;    	   /* is this an ogg stream or not */
	unsigned int   ogg_stream_serial;  /* The serial number of the stream */

};


/* This is global because both compression.c and vadclient.c need to write */ 
struct ringbuffer  {

	unsigned int    used;
	unsigned int    size;

	unsigned char*  outp;
	unsigned char*  inp;

	unsigned char*  start;
	unsigned char*  end;

};


/* This is global because we pass this information to the compression code */
struct net_audio_params_pack  {

	char id[8];
	unsigned long long   	timestamp;
	unsigned long  		end_of_stream;
	unsigned long  		sample_rate;
	unsigned long  		precision;
	unsigned long  		channels;

	/* Ogg Vorbis Headers, for decoding */
	char 	       ogg_flag;
	unsigned int   serialno;
	ogg_packet     headers[3];

	/* If this is an ogg stream the headers structures are followed by */
	/* the actual header data.					   */

};

/* XDR Routine for ogg_packet structure */
static int xdr_ogg_packet(xdrs, op)
	XDR *xdrs;
	ogg_packet *op;
{

	/* If this is an ogg stream we must broadcast the headers too.  */
	/* They are stored in the three ogg_packet structures. However  */
	/* we can't just send over ogg_packet structures beacuse they   */
	/* are simply pointers to buffers of data in memory. So we      */
	/* append the real header data after each header structure.     */

	/* Tf every routine works when encoding the fields */
	if(xdr_long(xdrs, &op->bytes)                  &&
	   xdr_long(xdrs, &op->b_o_s)                  &&
	   xdr_long(xdrs, &op->e_o_s)                  &&
	   xdr_long_long(xdrs, &op->granulepos) &&
	   xdr_long_long(xdrs, &op->packetno)   &&
	   xdr_opaque(xdrs, (char*)op->packet, op->bytes) ) 
		return(1);


	/* If any of the primitives fails, we fail */
	return(0);

}

/* XDR Routine for net_audio_params_pack */
static int xdr_net_audio_params_pack(xdrs, np)
	XDR *xdrs;
	struct net_audio_params_pack *np;
{
	if(xdr_long_long(xdrs, (char*)&np->id, 8)       &&
	   xdr_long_long(xdrs, &np->timestamp)   	&&
	   xdr_u_long(xdrs, &np->end_of_stream) 	&&
	   xdr_u_long(xdrs, &np->sample_rate)           &&
	   xdr_u_long(xdrs, &np->precision)             &&
	   xdr_u_long(xdrs, &np->channels)              &&
	   xdr_u_int(xdrs, &np->serialno)		&&
	   xdr_u_char(xdrs, &np->ogg_flag))  {

		/* If this is an ogg stream, get the headers too */
		if(np->ogg_flag == 1) {
		
			if(xdr_ogg_packet(xdrs, &np->headers[0]) &&
			   xdr_ogg_packet(xdrs, &np->headers[1]) &&
			   xdr_ogg_packet(xdrs, &np->headers[2])   )
				return(1);
		
		}

		/* Otherwise, we still have a good packet */
		else
			return(1);

	}

	/* If any routine failed, the whole thing has failed */
	return(0);  /* Failure */

}



/* structure for the header of the audio_data_packet */
struct audio_data_packet_hdr {

	/* Length of the data portion of our packet */
	unsigned long len;

	/* timestamp is a very important parameter of the audio_data_packet.  */
	/* This is not a timestamp of when the packet was sent, it is a time  */
	/* at which this packet should be played. The clients used this value */
	/* along with the current time that is sent with each net params pack */
	/* to decide whether they are going to slow or two fast.              */
	unsigned long long timestamp;

};

/* XDR Routine to read and convert audio_data_packet_hdr's */
static int xdr_audio_data_packet_hdr(xdrs, hdr) 
	XDR *xdrs;
	struct audio_data_packet_hdr *hdr;
{
	if(xdr_u_long(xdrs, &hdr->len)			&&
	   xdr_long_long(xdrs, &hdr->timestamp))
		return(1);

	return(0);	

}

/* This is the format of the audio data packet */
struct audio_data_packet  {

	/* Header */
	struct audio_data_packet_hdr hdr;

	/* This buffer holds the audio data. We probably won't use it all so  */
	/* so when we send the packet we WONT specify the size of the message */
	/* with something like sizeof(struct audio_data_packet).              */
	char data[MAX_BUFSIZE];

};

/* XDR Routine to read and convert audio_data_packet's */
static int xdr_audio_data_packet(xdrs, ap) 
	XDR *xdrs;
	struct audio_data_packet *ap;
{
	if(xdr_audio_data_packet_hdr(xdrs, &ap->hdr)	    &&
	   xdr_opaque(xdrs, (char*)&ap->data, ap->hdr.len))
		return(1);

	return(0);	

}

/* timestamp of the last net packet. This time is updated locally between net */
/* packets. The net packet timestamp represents a virtual time for the server */
/* that is passed to clients. It is updated on one second intervals, however  */
/* we will count in milliseconds between arrivals of succesive packets.       */
unsigned long long server_time;		

/* This is the time in terms of audio played so far */
unsigned long long audio_time;


/* Writes our data to the ring buffer */
extern int write_to_buffer(char *buffer, int size, unsigned long long timestamp);

/* Close everything thing up and die when we get an error */
extern void error_exit(char *message);

/* Defines for Big and Little Endian */
#define _BENDIAN_  0
#define _LENDIAN_  1

/* Test for byte order */
/* Taken from, http://codeproject.com/cpp/endianess.asp */
static int TestByteOrder()
{
	short int word = 0x0001;
	char *byte = (char *) &word;
	return(byte[0] ? _LENDIAN_ : _BENDIAN_);
}

/* Function to swap the bytes of a 16 bit number */
static inline uint16_t SWAP16(uint16_t word) 
{
	return(((word>>8)) | (word << 8));
}

/* We had to make this simple routine to go from little endian to host order */
/* because linux doesn't seem to have a routine to do it. We use this to     */
/* convert the audio samples.						     */
static inline uint16_t LETOH16(uint16_t word) 
{
	if(TestByteOrder() == _LENDIAN_) 
		return(word);
	else
		return(SWAP16(word));
}


#endif
