#include <stdio.h>
#include <sys/types.h> 
#include <sys/socket.h> 
#include <netinet/in.h> 
#include <arpa/inet.h> 
#include <vorbis/vorbisenc.h>

#include "vadserv.h"			/* Defines and structs needed application wide */
#include "compression.h"

/* Ogg Vorbis Encoding Variables, most of this code is just a direct copy of the example in the vorbis module */
ogg_stream_state os; /* take physical pages, weld into a logical stream of packets */
			  
ogg_page         og; /* one Ogg bitstream page.  Vorbis packets are inside */
ogg_packet       op; /* one raw packet of data for decode */
  
vorbis_info      vi; /* struct that stores all the static vorbis bitstream
			  settings */
vorbis_comment   vc; /* struct that stores all the user comments */

vorbis_dsp_state vd; /* central working state for the packet->PCM decoder */
vorbis_block     vb; /* local working space for packet->PCM decode */


/* FLAGS */
int vorbis_setup=0;  /* Marks whether vorbis has been set up yet */
int reset_used=0;    /* Marks to reset unecnoded_used to 0 */
int new_stream=0;    /* Marks whether encode has been called for the stream */

/* Global object of newly encoded data, we give a pointer to this struct when */
/* when we have enough data to get an ogg page and it needs to be sent.       */
struct encoded_audio_data encoded_data;

/* set  up the encoding evironment */
int encode_init(struct audio_params *format, struct net_audio_params_pack *net_packet, float quality)

{
	
	/* Three headers */
	ogg_packet headers[3];

	int ret, i, serialno;
	
	/* Initializes the vorbis info structure, must call a vorbis_info_clear when done */
	vorbis_info_init(&vi);

	/* Set the encoding mode, pass the number of channels, sample rate, and quality mode */
  	ret=vorbis_encode_init_vbr(&vi, format->channels, format->sample_rate, quality);

	/* Error: unsupported format */	
	if(ret)
		return ret;	

	/* add a comment */
  	vorbis_comment_init(&vc);
  	vorbis_comment_add_tag(&vc,"ENCODER","VAD Server");

	/* set up the analysis state and auxiliary encoding storage */
	vorbis_analysis_init(&vd,&vi);
  	vorbis_block_init(&vd,&vb);

	/* set up the packet-stream encoder, use rand for the serial no */
	srand(time(NULL));
	serialno = rand();
	ogg_stream_init(&os, serialno);
	net_packet->serialno = serialno;
	
	
	/* Here we set up the three headers for the ogg stream */
	/* Since these are needed to decode we want to re-broadcast */
	/* them so that clients can pick up in the middle of a stream */
	vorbis_analysis_headerout(&vd, &vc, &headers[0], &headers[1], &headers[2]);
	
	/* We need to copy the header packets into our net_audio_params_pack */
	/* Do it for all three */	
	for(i=0;i<3;i++)  {
	
		/* Copy the data to the packet, allocate new space for the actaul packet data */
		memcpy(&net_packet->headers[i], &headers[i], sizeof(ogg_packet));
		net_packet->headers[i].packet = (unsigned char *)malloc(headers[i].bytes);
		memcpy(net_packet->headers[i].packet, headers[i].packet, headers[i].bytes); 
		
		/* Submit the header to the stream, even though we are broadcasting it to */
		/* I figure it is better this way since the stream is still to spec. With */
		/* the headers at the front of the stream.				  */
		ogg_stream_packetin(&os, &headers[i]);
	
	}
			

	/* As per ogg vorbis spec, the first encoded pages that come out the */
	/* engine will be the headers that we have just submitted. To keep   */
	/* in line with the specification we will flush the headers from the */
	/* stream and into the encoded data buffer so that they can be sent. */
	
	/* New page for the headers */
	encoded_data.encoded_used = 0;
	
	
	/* While there are still header pages in the stream */	
	while( ogg_stream_flush(&os, &og) != 0 ) {  
	
		/* Copy the header(of this header page) into our data buffer */	
		memcpy(encoded_data.data+encoded_data.encoded_used, og.header, og.header_len);
		encoded_data.encoded_used+=og.header_len;
		/* Then copy the body of the page into our data buffer.      */	
		memcpy(encoded_data.data+encoded_data.encoded_used, og.body, og.body_len);
		encoded_data.encoded_used+=og.body_len;
			
	}

	/* Headers are not compressed so unencoded_used = encoded_used */
	encoded_data.unencoded_used = encoded_data.encoded_used;
	
	/* Store format info for the data as well */
	encoded_data.sample_rate = format->sample_rate;
	encoded_data.precision   = format->precision;
	encoded_data.channels    = format->channels;
	
	/* We want to return how much data the headers use */
	ret = encoded_data.encoded_used;
	
	/* Mark vorbis as set up */
	vorbis_setup=1;	
	
	/* Mark this as a new stream */
	new_stream = 1;
	
	/* Before we return, mark unencode_used as needing to be reset. Since */
	/* the next call of encode will count as a new page. We want it to    */
	/* to start at 0.						      */
	reset_used=1;
	
	/* We achieved success, ready to encode data */
	return ret;	
	
}




/* DeInitializes all of the encoding setup */
void encode_deinit(struct net_audio_params_pack *net_packet)

{
	
	int i;
	
	/* Free the buffers allocated for our copies of the headers */
	for(i=0;i<3;i++)  
		free(net_packet->headers[i].packet);
	
	/* clean up.  vorbis_info_clear() must be called last */
  	ogg_stream_clear(&os);
  	vorbis_block_clear(&vb);
  	vorbis_dsp_clear(&vd);
  	vorbis_comment_clear(&vc);
  	vorbis_info_clear(&vi);

}






int encode(char *buffer, int size)

{

	int eos=0;				/* End of stream */
	int total=0, nw=0;
	
	if(size <= 0)
		return -1; 

	/* Marks the end of the stream */
	/*
	if(size == 0) 
		vorbis_analysis_wrote(&vd, 0);
	*/
	
	else {
	
		float **encode_buffer = vorbis_analysis_buffer(&vd, size);
		int i;

		/* Uninterleave the samples, this code is straight from encoder_example.c */
		/* in the vorbis examples that come with the source			  */
		for(i=0;i<size/4;i++){
			encode_buffer[0][i]=((buffer[i*4+1]<<8)|
		      (0x00ff&(int)buffer[i*4]))/32768.f;
			encode_buffer[1][i]=((buffer[i*4+3]<<8)|
		      (0x00ff&(int)buffer[i*4+2]))/32768.f;
      		}
		
		/* Let them know how much we wrote */
		vorbis_analysis_wrote(&vd, i);	

	}	

	/* If this was a new stream, it isn't anymore */
	if(new_stream)
		new_stream=0;

	/* If we need to reset unencoded used do it */
	if(reset_used) {
		encoded_data.unencoded_used = 0;
		encoded_data.encoded_used = 0;		
		reset_used=0;
	}
		
	/* Update the number of unencoded audio bytes in this page */
	encoded_data.unencoded_used+=size;
	
	/* Now we start the actual encoding */
	while(vorbis_analysis_blockout(&vd, &vb) == 1)  {

		vorbis_analysis(&vb, NULL);
		vorbis_bitrate_addblock(&vb);
		
		while(vorbis_bitrate_flushpacket(&vd, &op))  {

			ogg_stream_packetin(&os, &op);	
	
			while(!eos)  {
				int ret = ogg_stream_pageout(&os, &og);
				int encoded_size; 
				
				/* If there aren't any pages then were out */
				if(ret == 0)  
					break;
				
								
				encoded_size = og.header_len + og.body_len;
				
				/* Ignore this page if it too big. Not a good idea.  */
				/* XXX This definitely needs to be fixed soon :) XXX */
				if(encoded_data.encoded_used + encoded_size > MAX_ENCODED_DATA_SIZE)  {
					printf("encode: dropped ogg_page\n"); 
					break;
				}
			
				/* Copy the page(header and body) into the buffer */	
				memcpy(encoded_data.data+encoded_data.encoded_used, og.header, og.header_len);
				encoded_data.encoded_used+=og.header_len;
				memcpy(encoded_data.data+encoded_data.encoded_used, og.body, og.body_len);
				encoded_data.encoded_used+=og.body_len;
	
				/* Add this page to the total number of bytes of encoded data */	
				total+=og.header_len+og.body_len;
		
				/* Reset unencoded_used on the next call of    */
				/* encode since we are starting a new page.    */
				reset_used=1;
				
				/* If this is the end of our stream, kind of pointelss really */
				if(ogg_page_eos(&og))  
					eos=1;
				
			} /* while(!eos) */
		
		} /* while(vorbis_bitrate_flushpacket(&vd, &op)) */
		 
	} /* while(vorbis_analysis_blockout(&vd, &vb) == 1) */
		
	/* Return the number of bytes the encoded_data uses */	
	return total;
	
}

/* Has vorbis been setup previouslly */
int is_vorbis_setup() { return vorbis_setup; }

/* Has the encode function been called at all for this stream */
int is_new_stream()   { return new_stream;   }

/* Return a pointer to the recently encoded data */
struct encoded_audio_data * get_encoded_data() { return(&encoded_data); }
