#include <sys/types.h> 
#include <sys/socket.h> 
#include <sys/ioctl.h>
#include <netinet/in.h> 
#include <arpa/inet.h> 
#include <rpc/rpc.h>	/* XDR */

#include <time.h>
#include <fcntl.h>
#include <unistd.h>
#include <time.h>
#include <stdio.h>
#include <math.h>
#include <getopt.h>
#include <vorbis/vorbisenc.h>

#include "vadserv.h"
#include "compression.h"		/* compression Routines */




#define OPTION_STRING  "a:b:d:"

static struct option long_options[] = 
{  
   { "address", required_argument, NULL, 'a' },
   { "bufsize", required_argument, NULL, 'b' },
   { "device",  required_argument, NULL, 'd' },
   { 0, 0, 0, 0 } 
};


/* Byte pattern to signify a net parameter packet */
char * net_byte_pattern = "NETPARAM";

char* group=NULL;          /* Address of the client, should be multicast soon */

int bufsize = 0;	   /* size used to sendto, packet size */ 	

int sock=0;		   /* File descriptor for the socket */
int svad=0;  		   /* File descriptor for svad device file */

struct sockaddr_in addr;   /* Address of clients, multicast */  

int total_audio_bytes_sent=0; 
int total_bytes_in_packet=0;

struct audio_params vad_params;	/* current audio format the VAD is using */ 



/* Get the encoding format, works for both the VAD in OpenBSD 
   and maudio in linux */
int get_encoding(int fd, char * buf, int size, struct audio_params *format);

/* Get the status of the virtual audio driver. This works in both OpenBSD and */
/* linux. This allows us to see if we can read data off of the slave side.    */
int get_vad_status(int svadfd);


/* Accepts and encoded_audio_data structure pointer, and a socket and an      */
/* address to send to. The function packetizes the audio data and then sends. */
int pack_and_send(struct encoded_audio_data *data, int socket, struct sockaddr_in* addr);

/* Timing Functions */
time_t start_t=0;	/* The time when we started streaming */

/* get_time() returns the number of seconds passed since streaming has started */
int get_time() { return(time(NULL) - start_t); }

/* Initializes the timer system, at the beginning of a new stream */
void init_time()  { start_t = time(NULL); }





int main(int argc, char *argv[]) 

{ 

	char ch;		/* command line option */
	
	int nw=0;      		/* Number of bytes written on the socket */
	int nr=0;		/* Number of bytes read on svad */
	
	char message[MAX_BUFSIZE]; /* Buffer for audio data */
	
	/* A packet sent to the client that describes audio stream format */
	struct net_audio_params_pack net_packet; 
	
	int bytes_per_sec;	/* # of bytes in one second of audio data */
	int encoding_changed=0;	/* Has the encoding has changed? */

	printf("VAD - Virtual Audio Driver - Server\n\n");
	
	while((ch = getopt_long(argc, argv, OPTION_STRING, long_options, NULL)) != -1)  {
		switch(ch) {
			
			case 'a':   /* Set the address */
				group = optarg;
				break;
			
			case 'b':   /* Change bufsize */
				bufsize = strtol(optarg, NULL, 10);
				
				if(bufsize > MAX_BUFSIZE)
					bufsize = MAX_BUFSIZE;

				if(bufsize < MIN_BUFSIZE)
					bufsize = MIN_BUFSIZE;
				break;
			
			case 'd':   /* Specify the device */
		
				if((svad = open(optarg, O_RDONLY, 0)) < 0) {
					perror("open svad device");
					exit(1);
				}
				
				break;
			
			case '?':
				return(-1);
				break;
		}	
	}

	if(group==NULL)  
		group = DEFAULT_ADDR;

	if(bufsize == 0)
		bufsize = DEFAULT_BSIZE;
		
	if(svad == 0)  {
		if((svad = open(DEFAULT_DEV, O_RDONLY, 0)) < 0) {
			perror("open device:");
			exit(1);
		}
	}
		
	/* Open the socket for sending */	
	if((sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		perror("socket"); 
		exit(1); 
	}

	bzero(&addr, sizeof(addr)); 
	addr.sin_family = AF_INET; 
	addr.sin_addr.s_addr = htonl(INADDR_ANY); 
	addr.sin_port = htons(PORT); 
	addr.sin_addr.s_addr = inet_addr(group); 

	
	
	while (1) { 

		
		/* Read the data off of the VAD slave */		
		if( (nr = read(svad, message, bufsize)) <= 0) {
			perror("read");
			close(sock);
			exit(1);
		}

		/* Get the encoding format of the audio data read off of the device */
		/* get_encoding returns 1 if it the encoding has been changed and   */
		/* dicovered. If it returns 0 it means the encoding hasn't changed  */
		/* since last time, so we can ignore it.			    */
		encoding_changed = get_encoding(svad, message, nr, &vad_params);	

		/* There was an error, exit */
		if(encoding_changed < 0) {
			printf("get_encoding: failed with %d\n", encoding_changed);
			break;
		}
		
		/* If we have a new encoding then we must setup for a new stream */
		if(encoding_changed == 1) {

			int j; 	    /* simple counter */
			
			printf("Encoding change:\n");
			printf("rate = %d ; channels = %d ; bits = %d\n", 
			       vad_params.sample_rate, vad_params.channels,
			       vad_params.precision);
		
			/* This identifies the data following it as a parameter packet */
			for(j=0;j<8;j++)	
				net_packet.id[j] = net_byte_pattern[j];


			/* Put format information into the net audio control packet */
			net_packet.sample_rate = vad_params.sample_rate;
			net_packet.precision   = vad_params.precision;
			net_packet.channels    = vad_params.channels;
		
			/* We have a new stream so we need to reset the total bytes of audio data sent*/
			total_audio_bytes_sent=0;
				
			/* We have a new stream, initialize the timer to zero again */
			init_time();

			/* Set the net_packet's timestamp to zero */
			net_packet.timestamp = 0;

			/* calculate how many bytes are in one second of audio data */
			bytes_per_sec = vad_params.sample_rate * (vad_params.precision / 8) * vad_params.channels;
				
			/* If this is CD quality audio */
			/* We set up the vorbis encoding engine here */
			/* We do this for every new stream we get. Maybe we will support other formats later */
			/* XXX If there are not two channels in the audio data, then we dont encode either. XXX */
			/* XXX HAVE TO FIX THIS SO THAT WE CAN SUPPORT SINGLE CHANNEL ENCODING XXX */
	
			if(vad_params.sample_rate == 44100 && vad_params.precision == 16 && 
			  vad_params.channels == 2) {
	
				struct encoded_audio_data *headers;
				int ret;
					
				/* We only end the strean and call encode_deinit if vorbis has been setup */
				if(is_vorbis_setup() == 1) {
						
					/* This marks the end of the stream */
					encode(NULL, 0);
					encode_deinit(&net_packet);
				}
				
				ret = encode_init(&vad_params, &net_packet, .9);
				
				if(ret < 0)  {
					printf("Error: encode_init\n");
					return ret;	
				}
				
				/* Send the headers out as per the ogg vorbis */
				/* stream spec. Use get_encoded_data().       */ 	 	
				headers = get_encoded_data();
		     	
/*	
				nw = sendto(sock, headers->data, headers->encoded_used, 0,
					(struct sockaddr *)&addr, sizeof(addr));

				if (nw < 0) { 
		        		perror("main: sendto headers");
					exit(1);
				}      
*/
				/* Mark this stream as ogg vorbis */			       	
				net_packet.ogg_flag = 1;
				
			}
			
			/* Otherwise, it is not ogg encoded */
			else 
				net_packet.ogg_flag = 0;
				      
		
		}

		
		
		/* Send the net_audio_params_pack every second */
		/* Or if we just started a new stream and it is ogg */

		if( get_time() > net_packet.timestamp || 
		  ( is_new_stream() && net_packet.ogg_flag == 1) 
		  || !get_vad_status(svad) ) {	
			
			unsigned char buf[MAX_BUFSIZE];	  			
			struct net_audio_params_pack pack;
			XDR *enxdrs;
			int size;

			
			/* Update the timestamp for this audio control packet */
			net_packet.timestamp = (unsigned long long)get_time(); 	
			
			/* The complement of the vad status will tell the the */
			/* client whether the end of the stream is coming.    */
			/* When the vad master closes the next read call will */
			/* block so we wont send any more parameter packets   */
			/* updating the server clock. This causes the client  */
			/* to stop playing audio data. We need to tell it     */
			/* the server doesn't care about synchronizing the    */
			/* stream anymore.				      */
			net_packet.end_of_stream = !get_vad_status(svad);
	


			
			/* Set up a stream to encode the net_packet   */
			/* into XDR. We only handle packets that are  */
			/* MAX_BUFSIZE large. Should fix this.	      */	
			enxdrs = (XDR *)malloc(sizeof(XDR));		
			xdrmem_create(enxdrs, buf, MAX_BUFSIZE, XDR_ENCODE);
			
			/* Encode the packet into the buffer */
			if(!xdr_net_audio_params_pack(enxdrs, &net_packet)) {
				printf("Error: XDR Encoding\n");		
				return(-1);
			}

			/* Get the size of the packet */	
			size = xdr_getpos(enxdrs);
			
			/* Destroy the stream */
			xdr_destroy(enxdrs); 
			free(enxdrs);
		 
			/* Send the packet */	
			nw = sendto(sock, buf, size, 0,
				(struct sockaddr *)&addr, sizeof(addr));
			/* Check for failure */	
			if (nw < 0) { 
				perror("main: sendto net_audio_params_pack");
				exit(1);
			}      

			
		}
		
	


		
	
		/* If the stream is ogg than we encode it here and send it out */
		if(net_packet.ogg_flag == 1)  {

			u_int16_t *p;
			int i, ret; 
	
			/* Submit audio data to the encoding engine */	
			ret = encode(message, nr);
			
			/* If we submitted enough data for a packet */	
			if(ret > 0) {	
		
				/* Get a pointer to the encoded data */
				struct encoded_audio_data *encoded_data = get_encoded_data();
					
				nw = pack_and_send(encoded_data, sock, &addr);
				
				if (nw < 0) { 
		        		perror("main: sendto encoded_data");
					exit(1);
				}      

				total_bytes_in_packet=0;
			
			}
			
			/* Total of uncompressed audio in this packet */
			total_bytes_in_packet+=nr;
			total_audio_bytes_sent+=nr;
		
		}

		/* Otherwise we just send the data out raw */
		else {	
			
			/* This is confusing. We make an encoded_audio_data  */
			/* structure for meesage even though it is not       */
			/* compressed. So encoded_used=unencoded_used. We do */
			/* this so we dont have to right another function    */
			/* for unencoded packetization and sending.           */
			struct encoded_audio_data data;
			data.encoded_used = data.unencoded_used = nr;
			memcpy(data.data, message, nr);
			data.sample_rate = vad_params.sample_rate;
			data.precision   = vad_params.precision;
			data.channels    = vad_params.channels;	
			
			/* message is unencoded */
			nw = pack_and_send(&data, sock, &addr);
	 
			if (nw < 0) { 
				perror("sendto"); 
				exit(1); 
			} 	
		
			total_audio_bytes_sent+=nr;	
			total_bytes_in_packet=nr;
		}

	
		
		
		/* if the total amount of data we have sent for this stream  */
		/* is greater than the amount that can possibly be played in */
		/* the time we have been sending, then we need to slow down. */		
		/* This small section of code limits the rate at which we    */
		/* send the audio data so that we dont overload the clients. */
		if(total_audio_bytes_sent > (bytes_per_sec * get_time())) {
			struct timespec treq, trem;	/* Used for timing */
			int data_diff = total_audio_bytes_sent - 
				        (bytes_per_sec * get_time());
		
			/* Calculate amount of nanoseconds we need to sleep */
			/* data_diff holds how much data we have sent in    */
			/* excess. We convert this amount of data into the  */
			/* time it would take to play and then sleep for    */
			/* that period.					    */	
			treq.tv_nsec = 1000000000 * ((float)data_diff / 
					             (float)bytes_per_sec);
	
			nanosleep(&treq, &trem);
		}
		

		
	} 

	close(svad);

	return(0);
	
}



/* Get the encoding format, works for both the VAD in OpenBSD and maudio in linux */
/* Return 1 if the format was changed, 0 if not					  */
int get_encoding(int fd, char * buf, int size, struct audio_params *format)

{

/* This code is for getting the format of the audio data from the slave device on the VAD */
/* It is slightly more complex than just polling the device.				  */	
#if defined (OPENBSD_VAD)

	/* Audio Control Packet sent by driver, lets us no when encoding has changed */
	struct audio_params_pack pack;
	
	/* We want to see if the audio data has an audio parameter packet in it */
	memcpy(&pack, buf, sizeof(struct audio_params_pack));
	
	/* Check all of the id fields to see if this is a audio param packet */	
	if(pack.id[0] == 'V' && pack.id[1] == 'A' && pack.id[2] == 'D' &&
           pack.id[3] == 'P' && pack.id[4] == 'A' && pack.id[5] == 'R' &&
	   pack.id[6] == 'A' && pack.id[7] == 'M') {

		/* We can remove it from the data now */
		memmove(buf, buf+sizeof(struct audio_params_pack), size-sizeof(struct audio_params_pack));
		size-=sizeof(struct audio_params_pack);
		
		/* Store the format in the audio_params structure passed as and argument */	
		format->sample_rate = pack.ec.sample_rate;
		format->precision   = pack.ec.precision;
		format->channels    = pack.ec.channels;
		
		/* Return 1, meaning the format has changed */
		return 1;
	
	}

#endif

/* With maudio we must query the device repeatedlly to see if the */
/* format has changed. Unlike the VAD it will not notify us 	  */
#if defined(LINUX_MAUDIO)

	int sample_rate, precision, channels;

	/* Get the precision */
	if(ioctl(fd, SOUND_PCM_READ_BITS, &precision) == -1) 
		return(-1);
	
	/* Convert the format codes to number of bits precision */
	if(precision == AFMT_S16_LE)
		precision = 16;
	
	else if(precision == AFMT_U8)
		precision = 8;
	
	/* Everything else is unsupported, return an error */
	else 	
		return(-2);
	
	
	/* Get the rate */
	if(ioctl(fd, SOUND_PCM_READ_RATE, &sample_rate) == -1) 
		return(-3);

	/* Get the number of channels */
	if(ioctl(fd, SOUND_PCM_READ_CHANNELS, &channels) == -1) 
		return(-4);

	/* If the format has changed at all then update it */
	if(format->sample_rate != sample_rate || format->channels != channels ||
	   format->precision != precision)  {
	
		format->sample_rate = sample_rate;
		format->channels = channels;
		format->precision = precision;

		
		/* Signal that the format has changed from the one passed in */
		return(1);
		
	}
	
	
#endif


	return(0);
	
}


/* Put and encoded_audio_data object into an audio_data_packet object and sent it out */
int pack_and_send(struct encoded_audio_data *data, int socket, struct sockaddr_in* addr)

{

	XDR *enxdrs;
	struct audio_data_packet packet;
	unsigned char buf[MAX_BUFSIZE];
	float bytes_per_sec;
	int packet_size;
	int nw=0;
	static unsigned long long t_stamp = SECS_TO_BUFFER * 1000000;
	static unsigned int old_time;

	/* Error, No empty buffers */
	if(data == NULL || data->encoded_used == 0 || data->unencoded_used == 0)
		return -1;
	
	/* Copy the contents of our data buffer into the packet structure */
	memcpy(packet.data, data->data, data->encoded_used);	

	/* Set the length of the data */
	packet.hdr.len = data->encoded_used;
	
	

	/* Now we want to give this packet a timestamp of when it should be   */
	/* played. This time is relative to the one we send with every        */
        /* net_audio_params_pack. We add SECS_TO_BUFFER worth of time to this */	
	/* projected timestamp so that the client will wait that many secs to */
	/* play so that it buffers data. 				      */
	
	/* If this is a brand new stream, then reset the cumulative timestamp */
	/* We want the client to buffer a certain amount of audio data on the */
	/* other side to absord network hiccups and such. It is obvious that  */
	/* if the time we stored last time this function was called is > than */
	/* the current time then we have reset the timer, it is a new stream. */	
	if(get_time() < old_time) 
		t_stamp = SECS_TO_BUFFER * 1000000;
	
	/* Get the old time so we can check it again next time */
	old_time = get_time();		
	
	bytes_per_sec = (data->sample_rate * (data->precision / 8) 
			* data->channels);

	/* Set the packet timestamp eqaul to the one we calculated last time  */
	/* this function was called.					      */
	packet.hdr.timestamp = t_stamp;			
	packet.hdr.timestamp = packet.hdr.timestamp;
	
	
	/* Calculate the next timestamp to the nearest micro second */	
	t_stamp+=(1000000.0 *
		     ((double)data->unencoded_used / (double)bytes_per_sec) 
		 );

	/* Now encode the packet into XDR form */
	enxdrs = (XDR *)malloc(sizeof(XDR));
	xdrmem_create(enxdrs, buf, MAX_BUFSIZE, XDR_ENCODE);

	/* If we fail to encode the packet */
	if(!xdr_audio_data_packet(enxdrs, &packet)) {
		printf("Error: XDR Fail: audio_data_packet\n");	
		return -1;
	}

	/* Get the size of our encoded packet */	
	packet_size = xdr_getpos(enxdrs);
	
	/* Destroy the XDR stream */
	xdr_destroy(enxdrs);
	free(enxdrs);

	
	
	/* Now all that is left is to send it */
	nw = sendto(socket, buf, packet_size , 0, (struct sockaddr*)addr, sizeof(*addr));
	
	/* Return the number of bytes sent, if error nw = -1, that is fine */
	return(nw);	
	
}


/* Get the status of our virtual audio driver, (ie. can we read from it yet) */
/* In OpenBSD with the VAD/SVAD we use an ioctl that returns 1 if the master */
/* side is open and being written to and 0 otherwise. In Linux with maudio   */
/* simply return 1 because maudio never blocks, it simply gives us silence.  */
/* The paramter is the file descriptor of the open slave device.	     */
int get_vad_status(int svadfd) 
{

/* If we are dealing with openbsd and vad, use the ioctl */
#if defined (OPENBSD_VAD)
	int status;
	if(ioctl(svadfd, SVAD_GETSTATUS, &status)) {
		printf("get_vad_status: ioctl error\n");
		return(0);
	}	
	else
		return(status);
#endif	

/* if we are dealing with linux and maudio, just return 1 */
#if defined (LINUX_MAUDIO)	
	return(1);
#endif	

}
